"""
    ulmo.usgs.ned.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides access to the `National Elevation Dataset` provided by the `United States Geological
    Survey`_ `National Map`_ system.

    .. _National Elevation Dataset: http://ned.usgs.gov/
    .. _United States Geological Survey: http://www.usgs.gov/
    .. _National Map: http://nationalmap.gov

"""
from __future__ import print_function
from past.builtins import basestring

from geojson import Feature, FeatureCollection, Polygon
import logging
import os
import requests
from tsgettoolbox.ulmo import util


# NED ftp url.
NED_FTP_URL = "ftp://rockyftp.cr.usgs.gov/vdelivery/Datasets/Staged/NED/<layer>/IMG/"

# default file path (appended to default ulmo path)
DEFAULT_FILE_PATH = "usgs/ned/"

layer_dict = {
    "Alaska 2 arc-second": "4f70aaece4b058caae3f8de9",
    "1 arc-second": "4f70aa71e4b058caae3f8de1",
    "1/3 arc-second": "4f70aa9fe4b058caae3f8de5",
    "1/9 arc-second": "4f70aac4e4b058caae3f8de7",
}

# configure logging
LOG_FORMAT = "%(message)s"
logging.basicConfig(format=LOG_FORMAT)
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def get_available_layers():
    """return list of available data layers
    """
    return list(layer_dict.keys())


def get_raster_availability(layer, bbox=None):
    """retrieve metadata for raster tiles that cover the given bounding box
    for the specified data layer.

    Parameters
    ----------
    layer : str
        dataset layer name. (see get_available_layers for list)
    bbox : (sequence of float|str)
        bounding box of in geographic coordinates of area to download tiles
        in the format (min longitude, min latitude, max longitude, max latitude)

    Returns
    -------
    metadata : geojson FeatureCollection
        returns metadata including download urls as a FeatureCollection
    """

    base_url = "https://www.sciencebase.gov/catalog/items"
    params = [
        ("parentId", layer_dict[layer]),
        ("filter", "tags=IMG"),
        ("max", 1000),
        ("fields", "webLinks,spatial,title"),
        ("format", "json"),
    ]

    if bbox:
        xmin, ymin, xmax, ymax = [float(n) for n in bbox]
        polygon = "POLYGON (({}))".format(
            ",".join(
                [
                    (repr(x) + " " + repr(y))
                    for x, y in [
                        (xmin, ymax),
                        (xmin, ymin),
                        (xmax, ymin),
                        (xmax, ymax),
                        (xmin, ymax),
                    ]
                ]
            )
        )
        params.append(
            (
                "filter",
                'spatialQuery={{wkt:"{}",relation:"{}"}}'.format(polygon, "intersects"),
            )
        )

    features = []
    url = base_url
    while url:
        r = requests.get(url, params)
        print("retrieving raster availability from %s" % r.url)
        params = []  # not needed after first request
        content = r.json()
        for item in content["items"]:
            feature = Feature(
                geometry=Polygon(_bbox2poly(item["spatial"]["boundingBox"])),
                id=item["id"],
                properties={
                    "name": item["title"],
                    "layer": layer,
                    "format": ".img",
                    "download url": [
                        x for x in item["webLinks"] if x["type"] == "download"
                    ][0]["uri"],
                },
            )
            features.append(feature)

        if content.get("nextlink"):
            url = content["nextlink"]["url"]
        else:
            break

    return FeatureCollection(features)


def get_raster(
    layer, bbox, path=None, update_cache=False, check_modified=False, mosaic=False
):
    """downloads National Elevation Dataset raster tiles that cover the given bounding box
    for the specified data layer.

    Parameters
    ----------
    layer : str
        dataset layer name. (see get_available_layers for list)
    bbox : (sequence of float|str)
        bounding box of in geographic coordinates of area to download tiles
        in the format (min longitude, min latitude, max longitude, max latitude)
    path : ``None`` or path
        if ``None`` default path will be used
    update_cache: ``True`` or ``False`` (default)
        if ``False`` and output file already exists use it.
    check_modified: ``True`` or ``False`` (default)
        if tile exists in path, check if newer file exists online and download if available.
    mosaic: ``True`` or ``False`` (default)
        if ``True``, mosaic and clip downloaded tiles to the extents of the bbox provided. Requires
        rasterio package and GDAL.

    Returns
    -------
    raster_tiles : geojson FeatureCollection
        metadata as a FeatureCollection. local url of downloaded data is in feature['properties']['file']
    """
    _check_layer(layer)

    raster_tiles = _download_tiles(
        get_raster_availability(layer, bbox), path=path, check_modified=check_modified
    )

    if mosaic:
        if path is None:
            path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

        util.mkdir_if_doesnt_exist(os.path.join(path, "by_boundingbox"))
        xmin, ymin, xmax, ymax = [float(n) for n in bbox]
        uid = util.generate_raster_uid(layer, xmin, ymin, xmax, ymax)
        output_path = os.path.join(path, "by_boundingbox", uid + ".tif")

        if os.path.isfile(output_path) and not update_cache:
            return output_path

        raster_files = [tile["properties"]["file"] for tile in raster_tiles["features"]]
        util.mosaic_and_clip(raster_files, xmin, ymin, xmax, ymax, output_path)
        return [output_path]

    return raster_tiles


def _check_layer(layer):
    """
    make sure the passed layer name is one of the handled options
    """

    if not layer in get_available_layers():
        err_msg = "The specified layer parameter ({})".format(layer)
        err_msg += "\nis not in the available options:"
        err_msg += "\n\t".join(get_available_layers())
        raise ValueError(err_msg)


def _download_features(feature_ids, path=None, check_modified=False):
    if path is None:
        path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

    if isinstance(feature_ids, basestring):
        feature_ids = [feature_ids]

    tiles = []
    tile_fmt = ".img"
    for feature_id in feature_ids:
        url = SCIENCEBASE_ITEM_URL % feature_id
        metadata = requests.get(url).json()
        layer = [a for a in list(layer_dict.keys()) if a in metadata["title"]][0]
        layer_path = os.path.join(path, layer_dict[layer])
        tile_urls = [
            link["uri"] for link in metadata["webLinks"] if link["type"] == "download"
        ]
        tiles.append(
            {
                "feature_id": feature_id,
                "tiles": util.download_tiles(
                    layer_path, tile_urls, tile_fmt, check_modified
                ),
            }
        )

    return tiles


def _bbox2poly(bbox):
    xmin = bbox["minX"]
    xmax = bbox["maxX"]
    ymin = bbox["minY"]
    ymax = bbox["maxY"]

    return [[(xmin, ymin), (xmin, ymax), (xmax, ymax), (xmax, ymin), (xmin, ymin)]]


def _download_tiles(tiles, path=None, check_modified=False):

    if path is None:
        path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

    for tile in tiles["features"]:

        metadata = tile["properties"]
        layer_path = os.path.join(path, layer_dict[metadata["layer"]])
        tile["properties"]["file"] = util.download_tiles(
            layer_path, metadata["download url"], metadata["format"], check_modified
        )[0]

    return tiles



from __future__ import absolute_import, division, print_function

import os
import networkx as nx

from math import log
from toolz import pluck
from .convert import convert, ooc_types
from .append import append
from .utils import cls_name


# Write out as dot
def dot_graph(filename="conversions"):
    # Edges from Convert
    dg = nx.DiGraph()
    for a, b in convert.graph.edges():
        cost = convert.graph.edge[a][b]["cost"]
        dg.add_edge(
            cls_name(a),
            cls_name(b),
            cost=cost,
            penwidth=max(log(1.0 / (cost + 0.06)), 1),
        )

    # Edges from Append
    for a, b in append.funcs:
        if b is not object and a != b:
            dg.add_edge(cls_name(b), cls_name(a), color="blue")

    # Color edges
    for n in convert.graph.nodes() + list(pluck(0, append.funcs)):
        if issubclass(n, tuple(ooc_types)):
            dg.node[cls_name(n)]["color"] = "red"

    # Convert to pydot
    p = nx.to_pydot(dg)

    p.set_overlap(False)
    p.set_splines(True)

    with open(filename + ".dot", "w") as f:
        f.write(p.to_string())

    os.system("neato -Tpdf %s.dot -o %s.pdf" % (filename, filename))
    print("Writing graph to %s.pdf" % filename)
    os.system("neato -Tpng %s.dot -o %s.png" % (filename, filename))
    print("Writing graph to %s.png" % filename)


if __name__ == "__main__":
    dot_graph()



"""
    ulmo.lcra.waterquality.core
    ~~~~~~~~~~~~~~~~~~~~~
    This module provides access to data provided by the `Lower Colorado
    River Authority`_ `Water Quality`_ web site.
    .. _Lower Colorado River Authority: http://www.lcra.org
    .. _Water Quality: http://waterquality.lcra.org/
"""
from bs4 import BeautifulSoup
import logging
from geojson import Point, Feature, FeatureCollection

# import unicode

from tsgettoolbox.ulmo import util

import dateutil

# import datetime
import os.path as op

LCRA_WATERQUALITY_DIR = op.join(util.get_ulmo_dir(), "lcra/waterquality")


log = logging.getLogger(__name__)

import requests

import pandas as pd

source_map = {
    "LCRA": "Lower Colorado River Authority",
    "UCRA": "Upper Colorado River Authority",
    "CRMWD": "Colorado River Municipal Water District",
    "COA": "City of Austin",
    "TCEQ": "Texas Commission on Environmental Quality",
}

real_time_sites = {
    "6977": "Matagorda 4SSW",
    "6985": "Matagorda 7 SW",
    "6990": "Matagorda 8 SSW",
    "6996": "Matagorda 9 SW",
}

# try:
#     import cStringIO as StringIO
# except ImportError:
#     import StringIO


def get_sites(source_agency=None):
    """Fetches a list of sites with location and available metadata.
    Parameters
    ----------
    source_agency : LCRA used code of the that collects the data. there
    are sites whose sources are not listed so this filter may not return
    all sites of a certain source. see
    ``source_map``.
    Returns
    -------
    sites_geojson : geojson FeatureCollection
    """
    sites_url = "http://waterquality.lcra.org/"
    response = requests.get(sites_url)
    lines = response.content.decode("utf-8").split("\n")
    sites_unprocessed = [
        line.strip().strip("createMarker").strip("(").strip(")").split(",")
        for line in lines
        if "createMarker" in line
    ]
    sites = [_create_feature(site_info) for site_info in sites_unprocessed]
    if source_agency:
        if not source_agency.upper() in source_map.keys():
            log.info("the source %s is not recognized" % source_agency)
            return {}
        else:
            sites = [
                site
                for site in sites
                if site["properties"]["source"] == source_map[source_agency.upper()]
            ]
    sites_geojson = FeatureCollection(sites)

    return sites_geojson


def get_historical_data(site_code, start=None, end=None, as_dataframe=False):
    """Fetches data for a site at a given date.
    Parameters
    ----------
    site_code: str
        The site code to fetch data for. A list of sites can be retrieved with
        ``get_sites()``
    date : ``None`` or date (see :ref:`dates-and-times`)
        The date of the data to be queried. If date is ``None`` (default), then
        all data will be returned.
    as_dataframe : bool
        This determines what format values are returned as. If ``False``
        (default), the values dict will be a dict with timestamps as keys mapped
        to a dict of gauge variables and values. If ``True`` then the values
        dict will be a pandas.DataFrame object containing the equivalent
        information.
    Returns
    -------
    data_dict : dict
        A dict containing site information and values.
    """

    if isinstance(site_code, (str)):
        pass
    elif isinstance(site_code, (int)):
        site_code = str(site_code)
    else:
        log.error(
            "Unsure of the site_code parameter type. \
                Try string or int"
        )
        raise

    waterquality_url = (
        "http://waterquality.lcra.org/parameter.aspx?qrySite=%s" % site_code
    )
    waterquality_url2 = "http://waterquality.lcra.org/events.aspx"

    initial_request = requests.get(waterquality_url)
    initialsoup = BeautifulSoup(initial_request.content, "html.parser")

    sitevals = [
        statag.get("value", None)
        for statag in initialsoup.findAll(id="multiple")
        if statag.get("value", None)
    ]

    result = _make_next_request(
        waterquality_url2, initial_request, {"multiple": sitevals, "site": site_code}
    )

    soup = BeautifulSoup(result.content, "html.parser")

    gridview = soup.find(id="GridView1")

    results = []

    headers = [head.text for head in gridview.findAll("th")]

    # uses \xa0 for blank

    for row in gridview.findAll("tr"):
        vals = [_parse_val(aux.text) for aux in row.findAll("td")]
        if len(vals) == 0:
            continue
        results.append(dict(zip(headers, vals)))

    data = _create_dataframe(results)

    if start and not data.empty:
        data = data.ix[util.convert_date(start) :]

    if end and not data.empty:
        data = data.ix[: util.convert_date(end)]

    if as_dataframe:
        return data
    else:
        return data.to_dict(orient="records")


def get_recent_data(site_code, as_dataframe=False):
    """fetches near real-time instantaneous water quality data for the LCRA
    bay sites.
    Parameters
    ----------
    site_code : str
        The bay site to fetch data for. see `real_time_sites`
    as_dataframe : bool
        This determines what format values are returned as. If ``False``
        (default), the values will be list of value dicts. If ``True`` then
        values are returned as pandas.DataFrame.
    Returns
    -------
    list of values or dataframe.
    """
    if site_code not in real_time_sites.keys():
        log.info("%s is not in the list of LCRA real time salinity sites" % site_code)
        return {}
    data_url = "http://waterquality.lcra.org/salinity.aspx?sNum=%s&name=%s" % (
        site_code,
        real_time_sites[site_code],
    )
    data = pd.read_html(data_url, header=0)[1]
    data.index = data["Date - Time"].apply(lambda x: util.convert_datetime(x))
    data.drop("Date - Time", axis=1, inplace=True)
    data = data.applymap(_nan_values)
    data.dropna(how="all", axis=0, inplace=True)
    data.dropna(how="all", axis=1, inplace=True)
    columns = dict([(column, _beautify_header(column)) for column in data.columns])
    data.rename(columns=columns, inplace=True)
    data = data.astype(float)

    if as_dataframe:
        return data
    else:
        return util.dict_from_dataframe(data)


def _nan_values(value):
    if value == -998.0 or value == "--":
        return pd.np.nan
    else:
        return value


def _beautify_header(str):
    return (
        str.replace(u"\xb0", "deg")
        .lower()
        .replace("(", "")
        .replace(")", "")
        .replace(u"%", u"percent")
        .replace(" ", "_")
        .replace(u"/", "per")
    )


def get_site_info(site_code):
    sites = get_sites()
    site = [
        site
        for site in sites["features"]
        if site_code == site["properties"]["site_code"]
    ]
    return site


def _create_dataframe(results):
    df = pd.DataFrame.from_records(results)
    df["Date"] = df["Date"].apply(util.convert_date)
    df.set_index(["Date"], inplace=True)
    df.dropna(how="all", axis=0, inplace=True)
    df.dropna(how="all", axis=1, inplace=True)
    return df


def _create_feature(site_info_list):
    geometry = Point(
        (float(site_info_list[0].strip()), float(site_info_list[1].strip()))
    )
    site_type_code = site_info_list[3].replace('"', "").strip()
    site_props = _parse_site_str(site_info_list[2])
    site_props["parameter"] = _get_parameter(site_type_code)
    site_props["source"] = _get_source(site_type_code)
    site_props["water_body"] = _get_water_body(site_type_code)
    site_props["real_time"] = _real_time(site_type_code)
    return Feature(geometry=geometry, properties=site_props)


def _extract_headers_for_next_request(request):
    payload = dict()
    for tag in BeautifulSoup(request.content, "html.parser").findAll("input"):
        tag_dict = dict(tag.attrs)
        if tag_dict.get("value", None) == "tabular":
            #
            continue
        # some tags don't have a value and are used w/ JS to toggle a set of checkboxes
        payload[tag_dict["name"]] = tag_dict.get("value")
    return payload


def _get_source(site_type_code):
    internal_source_abbr = {
        "LCLC": "LCRA",
        "LCUC": "UCRA",
        "LCCW": "CRMWD",
        "LCAU": "COA",
        "WCFO": "TCEQ",
    }
    if site_type_code not in internal_source_abbr.keys():
        return None
    return source_map.get(internal_source_abbr[site_type_code])


def _get_parameter(site_type_code):
    if site_type_code == "Salinity" or site_type_code == "Conductivity":
        return site_type_code
    else:
        return None


def _get_water_body(site_type_code):
    if site_type_code == "Bay":
        return "Bay"
    else:
        return None


def _make_next_request(url, previous_request, data):
    data_headers = _extract_headers_for_next_request(previous_request)
    data_headers.update(data)
    return requests.post(url, cookies=previous_request.cookies, data=data_headers)


def _parse_val(val):
    # the &nsbp translates to the following unicode
    if val == u"\xa0":
        return None
    else:
        return val


def _parse_site_str(site_str):
    site_code = (
        site_str.split("<br />")[0]
        .replace('"', "")
        .replace("Site", "")
        .replace("Number", "")
        .replace(":", "")
        .strip()
    )
    site_description = site_str.split("<br />")[1].strip('"')
    return dict(site_code=site_code, site_description=site_description)


def _real_time(site_type_code):
    if site_type_code == "Salinity" or site_type_code == "Conductivity":
        return True
    else:
        return False



from tsgettoolbox.odo.utils import (
    ext,
    iter_except,
    keywords,
    gentemp,
    records_to_tuples,
)


def test_ext():
    assert ext("a.py") == "py"
    assert ext("apy") == ""
    assert ext("myfile.csv.gz") == "gz"


def test_iter_except():
    s = set([1, 2, 3])

    def iterate():
        yield 1
        yield 2
        yield 4

    elems = iterate()
    inset = iter_except(lambda: next(elems) in s, StopIteration)
    assert list(inset) == [True, True, False]


def test_keywords():
    assert keywords(keywords) == ["func"]


def test_gentemp():
    i, _, data = next(gentemp([[1, 2], [3, 4]], start=1))
    assert i == 1 and data == [1, 2]


def test_records_to_tuples():
    seq = [{"i": 1, "f": 1.0}, {"i": 2, "f": 2.0}]
    lst = list(records_to_tuples("var * {i: int, f: float64}", seq))
    assert lst == [(1, 1.0), (2, 2.0)]


def test_records_to_tuples_mismatch_passthrough():
    assert records_to_tuples("var * int", "dummy") == "dummy"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import chain
import os
import os.path
import shutil
import sys

from setuptools import find_packages
from setuptools import setup

pkg_name = "tsgettoolbox"
version = open("VERSION").readline().strip()

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist")

    # The following block of code is to set the timestamp on files to
    # 'now', otherwise ChromeOS/google drive sets to 1970-01-01 and then
    # no one can install it because zip doesn't support dates before
    # 1980.
    os.chdir("dist")
    os.system("tar xvzf {pkg_name}-{version}.tar.gz".format(**locals()))
    os.system("find {pkg_name}-{version}* -exec touch {{}} \\;".format(**locals()))
    os.system(
        "tar czf {pkg_name}-{version}.tar.gz {pkg_name}-{version}".format(**locals())
    )
    shutil.rmtree("{pkg_name}-{version}".format(**locals()))
    os.chdir("..")

    os.system("twine upload dist/{pkg_name}-{version}.tar.gz".format(**locals()))
    sys.exit()

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.rst")).read()

install_requires = [
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
    "appdirs",
    "beautifulsoup4",
    "dask >= 0.11.1",
    "datashape >= 0.5.0",
    "future",
    "geojson",
    "html5lib<=0.9999999",
    "isodate",
    "lxml",
    "multipledispatch >= 0.4.7",
    "networkx >= 1.0",
    "numpy >= 1.7",
    "pandas >= 0.15.0",
    "pytest",
    "requests",
    "suds-jurko",
    "tables",
    "toolz >= 0.7.3",
    "tstoolbox >= 43.89.43.31",
    "zeep",
]


def read(filename):
    with open(filename, "r") as f:
        return f.read()


def read_reqs(filename):
    return read(filename).strip().splitlines()


def extras_require():
    extras = {
        req: read_reqs("etc/requirements_%s.txt" % req)
        for req in {
            "aws",
            "bcolz",
            "bokeh",
            "ci",
            "h5py",
            "mongo",
            "mysql",
            "postgres",
            "pytables",
            "sas",
            "ssh",
            "sql",
            "test",
        }
    }

    extras["mysql"] += extras["sql"]
    extras["postgres"] += extras["sql"]

    # don't include the 'ci' or 'test' targets in 'all'
    extras["all"] = list(
        chain.from_iterable(v for k, v in extras.items() if k not in {"ci", "test"})
    )
    return extras


setup(
    name="tsgettoolbox",
    version=version,
    description="Will get time series from different sources on the internet.",
    long_description=README,
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
    ],
    keywords="time_series uri url web_services rest",
    author="Tim Cera, P.E.",
    author_email="tim@cerazone.net",
    url="http://timcera.bitbucket.io/tsgettoolbox/docsrc/index.html",
    packages=find_packages(),
    package_data={"tsgettoolbox": ["tsgettoolbox/services/usgs/*.dat"]},
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    extras_require=extras_require(),
    entry_points={"console_scripts": ["tsgettoolbox=tsgettoolbox.tsgettoolbox:main"]},
    test_suite="tests",
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

try:
    import ConfigParser as cp
except ImportError:
    import configparser as cp

from tsgettoolbox import appdirs

dirs = appdirs.AppDirs("tsgettoolbox", "tsgettoolbox")


def read_api_key(service):
    # Read in API key
    if not os.path.exists(dirs.user_config_dir):
        os.makedirs(dirs.user_config_dir)
    configfile = os.path.join(dirs.user_config_dir, "config.ini")
    if not os.path.exists(configfile):
        with open(configfile, "w") as fp:
            fp.write(
                """

[{0}]
api_key = ReplaceThisStringWithYourKey

""".format(
                    service
                )
            )
    # Make sure read only by user.
    os.chmod(configfile, 0o600)

    inifile = cp.ConfigParser()
    inifile.readfp(open(configfile, "r"))

    try:
        api_key = inifile.get(service, "api_key")
    except BaseException:
        with open(configfile, "a") as fp:
            fp.write(
                """

[{0}]
api_key = ReplaceThisStringWithYourKey

""".format(
                    service
                )
            )
        api_key = "ReplaceThisStringWithYourKey"

    inifile.readfp(open(configfile, "r"))
    api_key = inifile.get(service, "api_key")
    if api_key == "ReplaceThisStringWithYourKey":
        raise ValueError(
            """
*
*   Need to edit {0}
*   to add your API key that you got from {1}.
*
""".format(
                configfile, service
            )
        )

    return api_key



"""
    ulmo.cuahsi.his_central.core
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module provides direct access to the `CUAHSI HIS Central`_ web service.


    .. _CUAHSI HIS Central: http://his.cuahsi.org/hiscentral.html
"""
from builtins import str
import suds.client

from tsgettoolbox.ulmo import util


HIS_CENTRAL_WSDL_URL = "http://hiscentral.cuahsi.org/webservices/hiscentral.asmx?WSDL"


def get_services(bbox=None):
    """Retrieves a list of services.


    Parameters
    ----------
    bbox : ``None`` or 4-tuple
        Optional argument for a bounding box that covers the area you want to
        look for services in. This should be a tuple containing (min_longitude,
        min_latitude, max_longitude, and max_latitude) with these values in
        decimal degrees. If not provided then the full set of services will be
        queried from HIS Central.


    Returns
    -------
    services_dicts : list
        A list of dicts that each contain information on an individual service.
    """
    suds_client = suds.client.Client(HIS_CENTRAL_WSDL_URL)

    if bbox is None:
        services = suds_client.service.GetWaterOneFlowServiceInfo()
    else:
        x_min, y_min, x_max, y_max = bbox
        services = suds_client.service.GetServicesInBox2(
            xmin=x_min, ymin=y_min, xmax=x_max, ymax=y_max
        )

    services = [_service_dict(service_info) for service_info in services.ServiceInfo]
    return services


def _cast_if_text(obj):
    """casts sax.text.Text objects to regular python strings, but leaves other
    objects unchanged
    """
    if isinstance(obj, suds.sax.text.Text):
        try:
            return str(obj)
        except UnicodeEncodeError:
            return str(obj)
    else:
        return obj


def _service_dict(service_info):
    """converts a ServiceInfo etree object into a service info dict"""
    change_keys = [
        # (old_key, new_key)
        ("aabstract", "abstract"),
        ("maxx", "max_x"),
        ("maxy", "max_y"),
        ("minx", "min_x"),
        ("miny", "min_y"),
        ("orgwebsite", "organization_website"),
        ("serv_url", "service_url"),
        ("sitecount", "site_count"),
        ("valuecount", "value_count"),
        ("variablecount", "variable_count"),
    ]

    service_dict = dict(
        [
            (util.camel_to_underscore(key), _cast_if_text(value))
            for key, value in dict(service_info).items()
        ]
    )

    for old_key, new_key in change_keys:
        if old_key in service_dict:
            service_dict[new_key] = service_dict[old_key]
            del service_dict[old_key]

    return service_dict



from __future__ import absolute_import, division, print_function

import numbers
import os
import re
import subprocess
import sys
import decimal
import warnings
from functools import partial

from operator import attrgetter
from itertools import chain
from collections import Iterator
from datetime import datetime, date, timedelta
from distutils.spawn import find_executable

import numpy as np
import pandas as pd
import sqlalchemy as sa
from sqlalchemy import inspect
from sqlalchemy.ext.compiler import compiles
from sqlalchemy import event
from sqlalchemy.schema import CreateSchema
from sqlalchemy.dialects import mssql, postgresql

from multipledispatch import MDNotImplementedError

import datashape
from datashape.dispatch import dispatch
from datashape.predicates import isdimension, isrecord, isscalar, isdatelike
from datashape import (
    DataShape,
    Record,
    Option,
    var,
    dshape,
    Map,
    discover,
    datetime_,
    date_,
    float64,
    int64,
    int_,
    string,
    bytes_,
    float32,
)

from toolz import (
    partition_all,
    keyfilter,
    valfilter,
    identity,
    concat,
    curry,
    merge,
    memoize,
)
from toolz.curried import pluck, map

from ..compatibility import unicode, StringIO
from ..directory import Directory
from ..utils import keywords, ignoring, iter_except, filter_kwargs, literal_compile
from ..convert import convert, ooc_types
from ..append import append
from ..resource import resource
from ..chunks import Chunks
from .csv import CSV

base = int, float, datetime, date, bool, str, decimal.Decimal, timedelta


# http://docs.sqlalchemy.org/en/latest/core/types.html

types = {
    "int64": sa.BigInteger,
    "int32": sa.Integer,
    "int": sa.Integer,
    "int16": sa.SmallInteger,
    "float32": sa.REAL,
    "float64": sa.FLOAT,
    "float": sa.FLOAT,
    "real": sa.FLOAT,
    "string": sa.Text,
    "date": sa.Date,
    "time": sa.Time,
    "datetime": sa.DateTime,
    "bool": sa.Boolean,
    "timedelta[unit='D']": sa.Interval(second_precision=0, day_precision=9),
    "timedelta[unit='h']": sa.Interval(second_precision=0, day_precision=0),
    "timedelta[unit='m']": sa.Interval(second_precision=0, day_precision=0),
    "timedelta[unit='s']": sa.Interval(second_precision=0, day_precision=0),
    "timedelta[unit='ms']": sa.Interval(second_precision=3, day_precision=0),
    "timedelta[unit='us']": sa.Interval(second_precision=6, day_precision=0),
    "timedelta[unit='ns']": sa.Interval(second_precision=9, day_precision=0),
    # ??: sa.types.LargeBinary,
}

revtypes = dict(map(reversed, types.items()))

# Subclass mssql.TIMESTAMP subclass for use when differentiating between
# mssql.TIMESTAMP and sa.TIMESTAMP.
# At the time of this writing, (mssql.TIMESTAMP == sa.TIMESTAMP) is True,
# which causes a collision when defining the revtypes mappings.
#
# See:
# https://bitbucket.org/zzzeek/sqlalchemy/issues/4092/type-problem-with-mssqltimestamp
class MSSQLTimestamp(mssql.TIMESTAMP):
    pass


# Assign the custom subclass as the type to use instead of `mssql.TIMESTAMP`.
mssql.base.ischema_names["TIMESTAMP"] = MSSQLTimestamp

revtypes.update(
    {
        sa.DATETIME: datetime_,
        sa.TIMESTAMP: datetime_,
        sa.FLOAT: float64,
        sa.DATE: date_,
        sa.BIGINT: int64,
        sa.INTEGER: int_,
        sa.BIGINT: int64,
        sa.types.NullType: string,
        sa.REAL: float32,
        sa.Float: float64,
        mssql.BIT: datashape.bool_,
        mssql.DATETIMEOFFSET: string,
        mssql.MONEY: float64,
        mssql.SMALLMONEY: float32,
        mssql.UNIQUEIDENTIFIER: string,
        # The SQL Server TIMESTAMP value doesn't correspond to the ISO Standard
        # It is instead just a binary(8) value with no relation to dates or times
        MSSQLTimestamp: bytes_,
    }
)

# Types which can be specified on precision.
# These are checked before checking membership in revtypes, because:
# 1) An instance of a precision type does not equal another instancec with
# the same precision.
# (DOUBLE_PRECISION(precision=53) != DOUBLE_PRECISION(precision=53)
# 2) Precision types can be a instance of a type in revtypes.
# isinstance(sa.Float(precision=53), sa.Float)
precision_types = {sa.Float, postgresql.base.DOUBLE_PRECISION}


def precision_to_dtype(precision):
    """
    Maps a float or double precision attribute to the desired dtype.

    The mappings are as follows:
    [1, 24] -> float32
    [25, 53] -> float64

    Values outside of those ranges raise a ``ValueError``.

    Parameter
    ---------
    precision : int
         A double or float precision. e.g. the value returned by
    `postgresql.base.DOUBLE_PRECISION(precision=53).precision`

    Returns
    -------
    dtype : datashape.dtype (float32|float64)
         The dtype to use for columns of the specified precision.
    """
    if isinstance(precision, numbers.Integral):
        if 1 <= precision <= 24:
            return float32
        elif 25 <= precision <= 53:
            return float64
    raise ValueError("{} is not a supported precision".format(precision))


# interval types are special cased in discover_typeengine so remove them from
# revtypes
revtypes = valfilter(lambda x: not isinstance(x, sa.Interval), revtypes)


units_of_power = {0: "s", 3: "ms", 6: "us", 9: "ns"}

# these aren't loaded by sqlalchemy by default
sa.dialects.registry.load("oracle")
sa.dialects.registry.load("postgresql")


def getbind(t, bind):
    if bind is None:
        return t.bind

    if isinstance(bind, sa.engine.interfaces.Connectable):
        return bind
    return create_engine(bind)


def batch(sel, chunksize=10000, bind=None):
    """Execute `sel`, streaming row at a time and fetching from the database in
    batches of size `chunksize`.

    Parameters
    ----------
    sel : sa.sql.Selectable
        Selectable to execute
    chunksize : int, optional, default 10000
        Number of rows to fetch from the database
    """

    def rowiterator(sel, chunksize=chunksize):
        with getbind(sel, bind).connect() as conn:
            result = conn.execute(sel)
            for rows in iter_except(
                curry(result.fetchmany, size=chunksize), sa.exc.ResourceClosedError
            ):
                if rows:
                    yield rows
                else:
                    return

    columns = [col.name for col in sel.columns]
    iterator = rowiterator(sel)
    return columns, concat(iterator)


@discover.register(sa.dialects.postgresql.base.INTERVAL)
def discover_postgresql_interval(t):
    return discover(sa.Interval(day_precision=0, second_precision=t.precision))


@discover.register(sa.dialects.oracle.base.INTERVAL)
def discover_oracle_interval(t):
    return discover(t.adapt(sa.Interval))


@discover.register(sa.sql.type_api.TypeEngine)
def discover_typeengine(typ):
    if isinstance(typ, sa.Interval):
        if typ.second_precision is None and typ.day_precision is None:
            return datashape.TimeDelta(unit="us")
        elif typ.second_precision == 0 and typ.day_precision == 0:
            return datashape.TimeDelta(unit="s")

        if typ.second_precision in units_of_power and not typ.day_precision:
            units = units_of_power[typ.second_precision]
        elif typ.day_precision > 0:
            units = "D"
        else:
            raise ValueError(
                "Cannot infer INTERVAL type with parameters"
                "second_precision=%d, day_precision=%d"
                % (typ.second_precision, typ.day_precision)
            )
        return datashape.TimeDelta(unit=units)
    if type(typ) in precision_types and typ.precision is not None:
        return precision_to_dtype(typ.precision)
    if typ in revtypes:
        return dshape(revtypes[typ])[0]
    if type(typ) in revtypes:
        return revtypes[type(typ)]
    if isinstance(typ, sa.Numeric):
        return datashape.Decimal(precision=typ.precision, scale=typ.scale)
    if isinstance(typ, (sa.String, sa.Unicode)):
        return datashape.String(typ.length, "U8")
    else:
        for k, v in revtypes.items():
            if isinstance(k, type) and (
                isinstance(typ, k) or hasattr(typ, "impl") and isinstance(typ.impl, k)
            ):
                return v
            if k == typ:
                return v
    raise NotImplementedError("No SQL-datashape match for type %s" % typ)


@discover.register(sa.ForeignKey, sa.sql.FromClause)
def discover_foreign_key_relationship(fk, parent, parent_measure=None):
    if fk.column.table is not parent:
        parent_measure = discover(fk.column.table).measure
    return {fk.parent.name: Map(discover(fk.parent.type), parent_measure)}


@discover.register(sa.sql.elements.ColumnClause)
def discover_sqlalchemy_column(c):
    meta = Option if getattr(c, "nullable", True) else identity
    return Record([(c.name, meta(discover(c.type)))])


@discover.register(sa.sql.FromClause)
def discover_sqlalchemy_selectable(t):
    ordering = {str(c): i for i, c in enumerate(c for c in t.columns.keys())}
    record = list(_process_columns(t.columns))
    fkeys = [
        discover(fkey, t, parent_measure=Record(record)) for fkey in t.foreign_keys
    ]
    for name, column in merge(*fkeys).items():
        index = ordering[name]
        _, key_type = record[index]
        # If the foreign-key is nullable the column (map) key
        # should be an Option type
        if isinstance(key_type, Option):
            column.key = Option(column.key)
        record[index] = (name, column)
    return var * Record(record)


def _process_columns(columns):
    """Process the dshapes of the columns of a table.

    Parameters
    ----------
    columns : iterable[column]
        The columns to process.

    Yields
    ------
    record_entry : tuple[str, dshape]
        A record entry containing the name and type of each column.
    """
    for col in columns:
        ((name, dtype),) = discover(col).fields
        yield str(name), dtype


@memoize
def metadata_of_engine(engine, schema=None):
    return sa.MetaData(engine, schema=schema)


def create_engine(uri, connect_args=None, **kwargs):
    """Creates a cached sqlalchemy engine.

    This differs from ``sa.create_engine``\s api by only accepting
    ``uri`` positionally.

    If the ``uri`` is an in memory sqlite database then this will not memioize
    the engine.
    """
    return (
        _create_engine_hashable_args
        if uri == "sqlite:///:memory:"
        else _memoized_create_engine_hashable_args
    )(uri, connect_args=frozenset((connect_args or {}).items()), **kwargs)


def _create_engine_hashable_args(uri, connect_args=None, **kwargs):
    """Unpacks non-hashable args for ``sa.create_engine`` and puts that back
    into whatever structure is expected.
    """
    return sa.create_engine(uri, connect_args=dict(connect_args or {}), **kwargs)


_memoized_create_engine_hashable_args = memoize(_create_engine_hashable_args)


@dispatch(sa.engine.base.Engine, str)
def discover(engine, tablename):
    metadata = sa.MetaData(engine)
    if tablename not in metadata.tables:
        try:
            metadata.reflect(engine, views=metadata.bind.dialect.supports_views)
        except NotImplementedError:
            metadata.reflect(engine)
    table = metadata.tables[tablename]
    return discover(table)


@dispatch(sa.engine.base.Engine)
def discover(engine):
    return discover(metadata_of_engine(engine))


@dispatch(sa.MetaData)
def discover(metadata):
    try:
        metadata.reflect(views=metadata.bind.dialect.supports_views)
    except NotImplementedError:
        metadata.reflect()
    pairs = []
    for table in sorted(metadata.tables.values(), key=attrgetter("name")):
        name = table.name
        try:
            pairs.append([name, discover(table)])
        except sa.exc.CompileError as e:
            warnings.warn(
                "Can not discover type of table {name}.\n"
                "SQLAlchemy provided this error message:\n\t{msg}"
                "\nSkipping.".format(name=name, msg=e.message),
                stacklevel=3,
            )
        except NotImplementedError as e:
            warnings.warn(
                "Odo does not understand a SQLAlchemy type.\n"
                "Odo provided the following error:\n\t{msg}"
                "\nSkipping.".format(msg="\n\t".join(e.args)),
                stacklevel=3,
            )
    return DataShape(Record(pairs))


@discover.register(sa.engine.RowProxy)
def discover_row_proxy(rp):
    return Record(list(zip(rp.keys(), map(discover, rp.values()))))


def validate_foreign_keys(ds, foreign_keys):
    # passed foreign_keys and column in dshape, but not a ForeignKey type
    for field in foreign_keys:
        if field not in ds.measure.names:
            raise TypeError(
                "Requested foreign key field %r is not a field in "
                "datashape %s" % (field, ds)
            )
    for field, typ in ds.measure.fields:
        if field in foreign_keys and not isinstance(getattr(typ, "ty", typ), Map):
            raise TypeError(
                "Foreign key %s passed in but not a Map "
                "datashape, got %s" % (field, typ)
            )

        if isinstance(typ, Map) and field not in foreign_keys:
            raise TypeError(
                "Map type %s found on column %s, but %r "
                "wasn't found in %s" % (typ, field, field, foreign_keys)
            )


def dshape_to_table(name, ds, metadata=None, foreign_keys=None, primary_key=None):
    """
    Create a SQLAlchemy table from a datashape and a name

    >>> dshape_to_table('bank', '{name: string, amount: int}') # doctest: +NORMALIZE_WHITESPACE
    Table('bank', MetaData(bind=None),
          Column('name', Text(), table=<bank>, nullable=False),
          Column('amount', Integer(), table=<bank>, nullable=False),
          schema=None)
    """

    if isinstance(ds, str):
        ds = dshape(ds)
    if not isrecord(ds.measure):
        raise TypeError(
            "dshape measure must be a record type e.g., "
            '"{a: int64, b: int64}". Input measure is %r' % ds.measure
        )
    if metadata is None:
        metadata = sa.MetaData()
    if foreign_keys is None:
        foreign_keys = {}

    validate_foreign_keys(ds, foreign_keys)

    cols = dshape_to_alchemy(ds, primary_key=primary_key or frozenset())
    cols.extend(
        sa.ForeignKeyConstraint([column_name], [referent])
        for column_name, referent in foreign_keys.items()
    )
    t = sa.Table(name, metadata, *cols, schema=metadata.schema)
    return attach_schema(t, t.schema)


@dispatch(object, str)
def create_from_datashape(o, ds, **kwargs):
    return create_from_datashape(o, dshape(ds), **kwargs)


@dispatch(sa.engine.base.Engine, DataShape)
def create_from_datashape(
    engine, ds, schema=None, foreign_keys=None, primary_key=None, **kwargs
):
    assert isrecord(ds), "datashape must be Record type, got %s" % ds
    metadata = metadata_of_engine(engine, schema=schema)
    for name, sub_ds in ds[0].dict.items():
        t = dshape_to_table(
            name,
            sub_ds,
            metadata=metadata,
            foreign_keys=foreign_keys,
            primary_key=primary_key,
        )
        t.create()
    return engine


def dshape_to_alchemy(dshape, primary_key=frozenset()):
    """

    >>> dshape_to_alchemy('int')
    <class 'sqlalchemy.sql.sqltypes.Integer'>

    >>> dshape_to_alchemy('string')
    <class 'sqlalchemy.sql.sqltypes.Text'>

    >>> dshape_to_alchemy('{name: string, amount: int}')
    [Column('name', Text(), table=None, nullable=False), Column('amount', Integer(), table=None, nullable=False)]

    >>> dshape_to_alchemy('{name: ?string, amount: ?int}')
    [Column('name', Text(), table=None), Column('amount', Integer(), table=None)]
    """
    if isinstance(dshape, str):
        dshape = datashape.dshape(dshape)
    if isinstance(dshape, Map):
        return dshape_to_alchemy(dshape.key.measure, primary_key=primary_key)
    if isinstance(dshape, Option):
        return dshape_to_alchemy(dshape.ty, primary_key=primary_key)
    if str(dshape) in types:
        return types[str(dshape)]
    if isinstance(dshape, datashape.Record):
        return [
            sa.Column(
                name,
                dshape_to_alchemy(getattr(typ, "ty", typ), primary_key=primary_key),
                primary_key=name in primary_key,
                nullable=isinstance(typ[0], Option),
            )
            for name, typ in dshape.parameters[0]
        ]
    if isinstance(dshape, datashape.DataShape):
        if isdimension(dshape[0]):
            return dshape_to_alchemy(dshape[1], primary_key=primary_key)
        else:
            return dshape_to_alchemy(dshape[0], primary_key=primary_key)
    if isinstance(dshape, datashape.String):
        fixlen = dshape[0].fixlen
        if fixlen is None:
            return sa.TEXT
        string_types = dict(U=sa.Unicode, A=sa.String)
        assert dshape.encoding is not None
        return string_types[dshape.encoding[0]](length=fixlen)
    if isinstance(dshape, datashape.DateTime):
        return sa.DATETIME(timezone=dshape.tz is not None)
    if isinstance(dshape, datashape.Decimal):
        return sa.NUMERIC(dshape.precision, dshape.scale)
    raise NotImplementedError("No SQLAlchemy dtype match for datashape: %s" % dshape)


@convert.register(Iterator, sa.Table, cost=300.0)
def sql_to_iterator(t, bind=None, **kwargs):
    _, rows = batch(sa.select([t]), bind=bind)
    return map(tuple, rows)


@convert.register(Iterator, sa.sql.Select, cost=300.0)
def select_to_iterator(sel, dshape=None, bind=None, **kwargs):
    func = pluck(0) if dshape and isscalar(dshape.measure) else map(tuple)
    _, rows = batch(sel, bind=bind)
    return func(rows)


@convert.register(base, sa.sql.Select, cost=200.0)
def select_to_base(sel, dshape=None, bind=None, **kwargs):
    with getbind(sel, bind).connect() as conn:
        return conn.execute(sel).scalar()


@append.register(sa.Table, Iterator)
def append_iterator_to_table(t, rows, dshape=None, bind=None, **kwargs):
    assert not isinstance(t, type)
    bind = getbind(t, bind)
    if not t.exists(bind=bind):
        t.create(bind=bind)
    rows = iter(rows)

    # We see if the sequence is of tuples or dicts
    # If tuples then we coerce them to dicts
    try:
        row = next(rows)
    except StopIteration:
        return t
    rows = chain([row], rows)
    if isinstance(row, (tuple, list)):
        dshape = dshape and datashape.dshape(dshape)
        if dshape and isinstance(dshape.measure, datashape.Record):
            names = dshape.measure.names
            if set(names) != set(discover(t).measure.names):
                raise ValueError(
                    "Column names of incoming data don't match "
                    "column names of existing SQL table\n"
                    "Names in SQL table: %s\n"
                    "Names from incoming data: %s\n"
                    % (discover(t).measure.names, names)
                )
        else:
            names = discover(t).measure.names
        rows = (dict(zip(names, row)) for row in rows)

    with bind.begin():
        for chunk in partition_all(1000, rows):  # TODO: 1000 is hardcoded
            bind.execute(t.insert(), chunk)

    return t


@append.register(sa.Table, Chunks)
def append_anything_to_sql_Table(t, c, **kwargs):
    for item in c:
        append(t, item, **kwargs)
    return t


@append.register(sa.Table, object)
def append_anything_to_sql_Table(t, o, **kwargs):
    return append(t, convert(Iterator, o, **kwargs), **kwargs)


@append.register(sa.Table, sa.Table)
def append_table_to_sql_Table(t, o, **kwargs):
    s = sa.select([o])
    return append(t, s, **kwargs)


@append.register(sa.Table, sa.sql.Select)
def append_select_statement_to_sql_Table(t, o, bind=None, **kwargs):
    t_bind = getbind(t, bind)
    o_bind = getbind(o, bind)
    if t_bind != o_bind:
        return append(t, convert(Iterator, o, bind=bind, **kwargs), bind=bind, **kwargs)
    bind = t_bind

    assert bind.has_table(t.name, t.schema), "tables must come from the same database"

    query = t.insert().from_select(o.columns.keys(), o)

    bind.execute(query)
    return t


def should_create_schema(
    ddl, target, bind, tables=None, state=None, checkfirst=None, **kwargs
):
    return ddl.element not in inspect(target.bind).get_schema_names()


def attach_schema(obj, schema):
    if schema is not None:
        ddl = CreateSchema(schema, quote=True)
        event.listen(
            obj,
            "before_create",
            ddl.execute_if(callable_=should_create_schema, dialect="postgresql"),
        )
    return obj


@resource.register(r"(.*sql.*|oracle|redshift)(\+\w+)?://.+")
def resource_sql(uri, *args, **kwargs):
    engine = create_engine(
        uri,
        # roundtrip through a frozenset of tuples so we can cache the dict
        connect_args=kwargs.pop("connect_args", {}),
        **filter_kwargs(sa.create_engine, kwargs)
    )
    ds = kwargs.pop("dshape", None)
    schema = kwargs.pop("schema", None)
    foreign_keys = kwargs.pop("foreign_keys", None)
    primary_key = kwargs.pop("primary_key", None)

    # we were also given a table name
    if args and isinstance(args[0], (str, unicode)):
        table_name, args = args[0], args[1:]
        metadata = metadata_of_engine(engine, schema=schema)

        with ignoring(sa.exc.NoSuchTableError):
            return attach_schema(
                sa.Table(table_name, metadata, autoload_with=engine), schema
            )
        if ds:
            t = dshape_to_table(
                table_name,
                ds,
                metadata=metadata,
                foreign_keys=foreign_keys,
                primary_key=primary_key,
            )
            t.create()
            return t
        else:
            raise ValueError("Table does not exist and no dshape provided")

    # We were not given a table name
    if ds:
        create_from_datashape(engine, ds, schema=schema, foreign_keys=foreign_keys)
    return engine


@resource.register("impala://.+")
def resource_impala(uri, *args, **kwargs):
    try:
        import impala.sqlalchemy
    except ImportError:
        raise ImportError("Please install or update `impyla` library")
    return resource_sql(uri, *args, **kwargs)


@resource.register("monetdb://.+")
def resource_monet(uri, *args, **kwargs):
    try:
        import monetdb
    except ImportError:
        raise ImportError("Please install the `sqlalchemy_monetdb` library")
    return resource_sql(uri, *args, **kwargs)


@resource.register("hive://.+")
def resource_hive(uri, *args, **kwargs):
    try:
        import pyhive
    except ImportError:
        raise ImportError("Please install the `PyHive` library.")

    pattern = "hive://((?P<user>[a-zA-Z_]\w*)@)?(?P<host>[\w.]+)(:(?P<port>\d*))?(/(?P<database>\w*))?"
    d = re.search(pattern, uri.split("::")[0]).groupdict()

    defaults = {"port": "10000", "user": "hdfs", "database": "default"}
    for k, v in d.items():
        if not v:
            d[k] = defaults[k]

    if d["user"]:
        d["user"] += "@"

    uri2 = "hive://%(user)s%(host)s:%(port)s/%(database)s" % d
    if "::" in uri:
        uri2 += "::" + uri.split("::")[1]

    return resource_sql(uri2, *args, **kwargs)


ooc_types.add(sa.Table)


@dispatch(sa.Table)
def drop(table, bind=None):
    bind = getbind(table, bind)
    table.drop(bind=bind, checkfirst=True)
    if table.exists(bind=bind):
        raise ValueError("table %r dropped but still exists" % table.name)
    metadata_of_engine(bind, schema=table.schema).remove(table)


@convert.register(sa.sql.Select, sa.Table, cost=0)
def table_to_select(t, **kwargs):
    return t.select()


@convert.register(pd.DataFrame, (sa.sql.Select, sa.sql.Selectable), cost=300.0)
def select_or_selectable_to_frame(el, bind=None, dshape=None, **kwargs):
    bind = getbind(el, bind)
    if bind.dialect.name == "postgresql":
        buf = StringIO()
        append(CSV(None, buffer=buf), el, bind=bind, **kwargs)
        buf.seek(0)
        datetime_fields = []
        other_dtypes = {}
        optional_string_fields = []
        try:
            fields = dshape.measure.fields
        except AttributeError:
            fields = [(0, dshape.measure)]

        for n, (field, dtype) in enumerate(fields):
            if isdatelike(dtype):
                datetime_fields.append(field)
            elif isinstance(dtype, Option):
                ty = dtype.ty
                if ty in datashape.integral:
                    other_dtypes[field] = "float64"
                else:
                    other_dtypes[field] = ty.to_numpy_dtype()
                    if ty == string:
                        # work with integer column indices for the
                        # optional_string columns because we don't always
                        # know the column name and then the lookup will fail
                        # in the loop below.
                        optional_string_fields.append(n)
            else:
                other_dtypes[field] = dtype.to_numpy_dtype()

        df = pd.read_csv(
            buf,
            parse_dates=datetime_fields,
            dtype=other_dtypes,
            skip_blank_lines=False,
            escapechar=kwargs.get("escapechar", "\\"),
        )
        # read_csv really wants missing values to be NaN, but for
        # string (object) columns, we want None to be missing
        columns = df.columns
        for field_ix in optional_string_fields:
            # use ``df.loc[bool, df.columns[field_ix]]`` because you cannot do
            # boolean slicing with ``df.iloc``.
            field = columns[field_ix]
            df.loc[df[field].isnull(), field] = None
        return df

    columns, rows = batch(el, bind=bind)
    dtypes = {}
    try:
        fields = dshape.measure.fields
    except AttributeError:
        fields = [(columns[0], dshape.measure)]
    for field, dtype in fields:
        if isinstance(dtype, Option):
            ty = dtype.ty
            try:
                dtypes[field] = ty.to_numpy_dtype()
            except TypeError:
                dtypes[field] = np.dtype(object)
            else:
                if np.issubdtype(dtypes[field], np.integer):
                    # cast nullable ints to float64 so NaN can be used for nulls
                    dtypes[field] = np.float64
        else:
            try:
                dtypes[field] = dtype.to_numpy_dtype()
            except TypeError:
                dtypes[field] = np.dtype(object)

    return pd.DataFrame(
        np.array(list(map(tuple, rows)), dtype=[(str(c), dtypes[c]) for c in columns])
    )


class CopyToCSV(sa.sql.expression.Executable, sa.sql.ClauseElement):
    def __init__(
        self,
        element,
        path,
        delimiter=",",
        quotechar='"',
        lineterminator="\n",
        escapechar="\\",
        header=True,
        na_value="",
        encoding=None,
        bind=None,
    ):
        self.element = element
        self.path = path
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.lineterminator = lineterminator
        self._bind = bind = getbind(element, bind)

        # mysql cannot write headers
        self.header = header and bind.dialect.name != "mysql"
        self.escapechar = escapechar
        self.na_value = na_value
        self.encoding = encoding

    @property
    def bind(self):
        return self._bind


try:
    from sqlalchemy.dialects.postgresql.psycopg2 import PGCompiler_psycopg2
except ImportError:
    pass
else:

    @partial(setattr, PGCompiler_psycopg2, "visit_mod_binary")
    def _postgres_visit_mod_binary(self, binary, operator, **kw):
        """Patched visit mod binary to work with literal_binds.

        When https://github.com/zzzeek/sqlalchemy/pull/366 is merged we can
        remove this patch.
        """
        literal_binds = kw.get("literal_binds", False)
        if getattr(self.preparer, "_double_percents", True) and not literal_binds:

            return "{} %% {}".format(
                self.process(binary.left, **kw), self.process(binary.right, **kw)
            )
        else:
            return "{} % {}".format(
                self.process(binary.left, **kw), self.process(binary.right, **kw)
            )


@compiles(CopyToCSV, "postgresql")
def compile_copy_to_csv_postgres(element, compiler, **kwargs):
    selectable = element.element
    if isinstance(selectable, sa.Table):
        selectable_part = compiler.preparer.format_table(selectable)
    else:
        selectable_part = "(%s)" % compiler.process(element.element, **kwargs)

    return "COPY %s TO STDOUT WITH (%s)" % (
        selectable_part,
        compiler.process(
            sa.text(
                """
                FORMAT CSV,
                HEADER :header,
                DELIMITER :delimiter,
                QUOTE :quotechar,
                NULL :na_value,
                ESCAPE :escapechar,
                ENCODING :encoding
                """
            ).bindparams(
                header=element.header,
                delimiter=element.delimiter,
                quotechar=element.quotechar,
                na_value=element.na_value,
                escapechar=element.escapechar,
                encoding=element.encoding
                or element.bind.execute("show client_encoding").scalar(),
            ),
            **kwargs
        ),
    )


@compiles(CopyToCSV, "mysql")
def compile_copy_to_csv_mysql(element, compiler, **kwargs):
    selectable = element.element
    return compiler.process(
        sa.text(
            """{0} INTO OUTFILE :path
            CHARACTER SET :encoding
            FIELDS TERMINATED BY :delimiter
            OPTIONALLY ENCLOSED BY :quotechar
            ESCAPED BY :escapechar
            LINES TERMINATED BY :lineterminator
            """.format(
                compiler.process(
                    selectable.select()
                    if isinstance(selectable, sa.Table)
                    else selectable,
                    **kwargs
                )
            )
        ).bindparams(
            path=element.path,
            encoding=element.encoding
            or element.bind.execute("select @@character_set_client").scalar(),
            delimiter=element.delimiter,
            quotechar=element.quotechar,
            escapechar=element.escapechar,
            lineterminator=element.lineterminator,
        )
    )


@compiles(CopyToCSV, "sqlite")
def compile_copy_to_csv_sqlite(element, compiler, **kwargs):
    if element.encoding is not None:
        raise ValueError(
            "'encoding' keyword argument not supported for " "SQLite to CSV conversion"
        )
    if not find_executable("sqlite3"):
        raise MDNotImplementedError("Could not find sqlite executable")

    # we are sending a SQL string directorly to the SQLite process so we always
    # need to bind everything before sending it
    kwargs["literal_binds"] = True

    selectable = element.element
    sql = (
        compiler.process(
            selectable.select() if isinstance(selectable, sa.Table) else selectable,
            **kwargs
        )
        + ";"
    )
    sql = re.sub(r"\s{2,}", " ", re.sub(r"\s*\n\s*", " ", sql)).encode(
        sys.getfilesystemencoding()  # we send bytes to the process
    )
    cmd = [
        "sqlite3",
        "-csv",
        "-%sheader" % ("no" if not element.header else ""),
        "-separator",
        element.delimiter,
        selectable.bind.url.database,
    ]
    with open(element.path, mode="at") as f:
        subprocess.Popen(cmd, stdout=f, stdin=subprocess.PIPE).communicate(sql)

    # This will be a no-op since we're doing the write during the compile
    return ""


try:
    from sqlalchemy_redshift.dialect import UnloadFromSelect
    from tsgettoolbox.odo.backends.aws import S3, get_s3_connection
except ImportError:
    pass
else:

    @resource.register("s3://.*/$")
    def resource_s3_prefix(uri, **kwargs):
        return Directory(S3)(uri, **kwargs)

    @append.register(Directory(S3), sa.Table)
    def redshit_to_s3_bucket(bucket, selectable, dshape=None, bind=None, **kwargs):
        s3_conn_kwargs = filter_kwargs(get_s3_connection, kwargs)
        s3 = get_s3_connection(**s3_conn_kwargs)

        unload_kwargs = filter_kwargs(UnloadFromSelect, kwargs)
        unload_kwargs["unload_location"] = bucket.path
        unload_kwargs["access_key_id"] = s3.access_key
        unload_kwargs["secret_access_key"] = s3.secret_key

        unload = UnloadFromSelect(selectable.select(), **unload_kwargs)

        with getbind(selectable, bind).begin() as conn:
            conn.execute(unload)
        return bucket.path


@append.register(CSV, sa.sql.Selectable)
def append_table_to_csv(csv, selectable, dshape=None, bind=None, **kwargs):
    kwargs = keyfilter(keywords(CopyToCSV).__contains__, merge(csv.dialect, kwargs))
    stmt = CopyToCSV(
        selectable,
        os.path.abspath(csv.path) if csv.path is not None else None,
        bind=bind,
        **kwargs
    )

    bind = getbind(selectable, bind)
    if bind.dialect.name == "postgresql":
        with csv.open("ab+") as f:
            with bind.begin() as conn:
                conn.connection.cursor().copy_expert(literal_compile(stmt), f)
    else:
        with bind.begin() as conn:
            conn.execute(stmt)
    return csv


try:
    from .hdfs import HDFS
except ImportError:
    pass
else:

    @append.register(HDFS(CSV), sa.sql.Selectable)
    def append_selectable_to_hdfs_csv(*args, **kwargs):
        raise MDNotImplementedError()



from __future__ import absolute_import
from __future__ import print_function

import logging
import os
from builtins import object
from io import BytesIO

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

import requests

from tstoolbox import tsutils

# EPA


class EPA_CSV(object):
    def __init__(self, url, **query_params):
        # Need to enforce waterml format
        query_params["mimeType"] = "csv"
        try:
            query_params["startDateLo"] = tsutils.parsedate(
                query_params["startDateLo"], strftime="%m-%d-%Y"
            )
        except KeyError:
            pass
        try:
            query_params["startDateHi"] = tsutils.parsedate(
                query_params["startDateHi"], strftime="%m-%d-%Y"
            )
        except KeyError:
            pass
        self.url = url
        self.query_params = query_params


@resource.register(r"https://www.waterqualitydata.us/data/Result/search", priority=17)
def resource_epa_csv(uri, **kwargs):
    return EPA_CSV(uri, **kwargs)


# Function to convert from EPA type to pd.DataFrame
@convert.register(pd.DataFrame, EPA_CSV)
def epa_csv_to_df(data, **kwargs):
    req = requests.get(data.url, params=data.query_params)
    if os.path.exists("debug_tsgettoolbox"):
        logging.warning(req.url)
    req.raise_for_status()

    ndf = pd.read_csv(BytesIO(req.content))
    return ndf


if __name__ == "__main__":
    r = resource(
        r"https://www.waterqualitydata.us/data/Result/search",
        characteristicName="Caffeine",
        bBox="-92.8,44.2,-88.9,46.0",
        startDateLo="10-01-2006",
        mimeType="csv",
    )

    as_df = odo(r, pd.DataFrame)
    print("Caffeine")
    print(as_df)



"""
    ulmo.usgs.eddn.core
    ~~~~~~~~~~~~~~~~~~~~~

    This module provides access to data provided by the `United States Geological
    Survey`_ `Emergency Data Distribution Network`_ web site.

    The `DCP message format`_ includes some header information that is parsed and
    the message body, with a variable number of characters. The format of the
    message body varies widely depending on the manufacturer of the transmitter,
    data logger, sensors, and the technician who programmed the DCP. The body can
    be simple ASCII, sometime with parameter codes and time-stamps embedded,
    sometimes not. The body can also be in 'Pseudo-Binary' which is character
    encoding of binary data that uses 6 bits of every byte and guarantees that
    all characters are printable.


    .. _United States Geological Survey: http://www.usgs.gov/
    .. _Emergency Data Distribution Network: http://eddn.usgs.gov/
    .. _http://eddn.usgs.gov/dcpformat.html

"""
from past.builtins import basestring

from bs4 import BeautifulSoup
from datetime import datetime, timedelta
import isodate
import logging
import os
import pandas as pd
import re
import requests
import shutil
from tsgettoolbox.ulmo import util

from . import parsers

# eddn query base url
EDDN_URL = "http://eddn.usgs.gov/cgi-bin/retrieveData.pl?%s"

# default file path (appended to default ulmo path)
DEFAULT_FILE_PATH = "usgs/eddn/"

# configure logging
LOG_FORMAT = "%(message)s"
logging.basicConfig(format=LOG_FORMAT)
log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def decode(dataframe, parser, **kwargs):
    """decodes dcp message data in pandas dataframe returned by ulmo.usgs.eddn.get_data().

    Parameters
    ----------
    dataframe : pandas.DataFrame
        pandas.DataFrame returned by ulmo.usgs.eddn.get_data()
    parser : {function, str}
        function that acts on dcp_message each row of the dataframe and returns a new dataframe
        containing several rows of decoded data. This returned dataframe may have different
        (but derived) timestamps than that the original row. If a string is passed then a matching
        parser function is looked up from tsgettoolbox.ulmo.usgs.eddn.parsers

    Returns
    -------
    decoded_data : pandas.DataFrame
        pandas dataframe, the format and parameters in the returned dataframe depend wholly on the parser used

    """
    if isinstance(parser, basestring):
        parser = getattr(parsers, parser)

    df = []
    for timestamp, data in dataframe.iterrows():
        parsed = parser(data, **kwargs)
        parsed.dropna(how="all", inplace=True)
        if not parsed.empty:
            df.append(parsed)

    df = pd.concat(df)
    return df


def get_data(
    dcp_address,
    start=None,
    end=None,
    networklist="",
    channel="",
    spacecraft="Any",
    baud="Any",
    electronic_mail="",
    dcp_bul="",
    glob_bul="",
    timing="",
    retransmitted="Y",
    daps_status="N",
    use_cache=False,
    cache_path=None,
    as_dataframe=True,
):
    """Fetches GOES Satellite DCP messages from USGS Emergency Data Distribution Network.

    Parameters
    ----------
    dcp_address : str, iterable of strings
        DCP address or list of DCP addresses to be fetched; lists will be joined by a ','.
    start : {``None``, str, datetime, datetime.timedelta}
        If ``None`` (default) then the start time is 2 days prior (or date of last data if cache is used)
        If a datetime or datetime like string is specified it will be used as the start date.
        If a timedelta or string in ISO 8601 period format (e.g 'P2D' for a period of 2 days) then
        'now' minus the timedelta will be used as the start.
        NOTE: The EDDN service does not specify how far back data is available. The service also imposes
        a maximum data limit of 25000 character. If this is limit reached multiple requests will be made
        until all available data is downloaded.
    end : {``None``, str, datetime, datetime.timedelta}
        If ``None`` (default) then the end time is 'now'
        If a datetime or datetime like string is specified it will be used as the end date.
        If a timedelta or string in ISO 8601 period format (e.g 'P2D' for a period of 2 days) then
        'now' minus the timedelta will be used as the end.
        NOTE: The EDDN service does not specify how far back data is available. The service also imposes
        a maximum data limit of 25000 character.
    networklist : str,
        '' (default). Filter by network.
    channel : str,
        '' (default). Filter by channel.
    spacecraft : str,
        East, West, Any (default). Filter by GOES East/West Satellite
    baud : str,
        'Any' (default). Filter by baud rate. See http://eddn.usgs.gov/msgaccess.html for options
    electronic_mail : str,
        '' (default) or 'Y'
    dcp_bul : str,
        '' (default) or 'Y'
    glob_bul : str,
        '' (default) or 'Y'
    timing : str,
        '' (default) or 'Y'
    retransmitted : str,
        'Y' (default) or 'N'
    daps_status : str,
        'N' (default) or 'Y'
    use_cache : bool,
        If True (default) use hdf file to cache data and retrieve new data on subsequent requests
    cache_path : {``None``, str},
        If ``None`` use default ulmo location for cached files otherwise use specified path. files are named
        using dcp_address.
    as_dataframe : bool
        If True (default) return data in a pandas dataframe otherwise return a dict.

    Returns
    -------
    message_data : {pandas.DataFrame, dict}
        Either a pandas dataframe or a dict indexed by dcp message times
    """

    if isinstance(dcp_address, list):
        dcp_address = ",".join(dcp_address)

    data = pd.DataFrame()

    if use_cache:
        dcp_data_path = _get_store_path(cache_path, dcp_address + ".h5")
        if os.path.exists(dcp_data_path):
            data = pd.read_hdf(dcp_data_path, dcp_address)

    if start:
        drs_since = _format_time(start)
    else:
        try:
            drs_since = _format_time(data["message_timestamp_utc"][-1])
        except:
            drs_since = "now -2 days"

    if end:
        drs_until = _format_time(end)
    else:
        drs_until = "now"

    params = {}
    params["DCP_ADDRESS"] = dcp_address
    params["DRS_SINCE"] = drs_since
    params["DRS_UNTIL"] = drs_until
    params["NETWORKLIST"] = networklist
    params["CHANNEL"] = channel
    params["BEFORE"] = ("//START\n",)
    params["AFTER"] = ("\n//END\n",)
    params["SPACECRAFT"] = spacecraft
    params["BAUD"] = baud
    params["ELECTRONIC_MAIL"] = electronic_mail
    params["DCP_BUL"] = dcp_bul
    params["GLOB_BUL"] = glob_bul
    params["TIMING"] = timing
    params["RETRANSMITTED"] = retransmitted
    params["DAPS_STATUS"] = daps_status

    data_limit_reached = True
    messages = []
    while data_limit_reached:
        new_message, data_limit_reached = _fetch_url(params)
        messages += new_message
        if data_limit_reached:
            params["DRS_UNTIL"] = _format_time(
                _parse(new_message[-1])["message_timestamp_utc"]
            )

    new_data = pd.DataFrame([_parse(row) for row in messages])

    if not new_data.empty:
        new_data.index = new_data.message_timestamp_utc
        data = new_data.combine_first(data)
        data.sort_index(inplace=True)

        if use_cache:
            # write to a tmp file and move to avoid ballooning h5 file
            tmp = dcp_data_path + ".tmp"
            data.to_hdf(tmp, dcp_address)
            shutil.move(tmp, dcp_data_path)

    if data.empty:
        if as_dataframe:
            return data
        else:
            return {}

    if start:
        if start.startswith("P"):
            start = data["message_timestamp_utc"][-1] - isodate.parse_duration(start)

        data = data[start:]

    if end:
        if end.startswith("P"):
            end = data["message_timestamp_utc"][-1] - isodate.parse_duration(end)

        data = data[:end]

    if not as_dataframe:
        data = data.T.to_dict()

    return data


def _fetch_url(params):
    r = requests.get(EDDN_URL, params=params)
    log.info("data requested using url: %s\n" % r.url)
    soup = BeautifulSoup(r.text, "lxml")
    message = soup.find("pre").contents[0].replace("\n", "").replace("\r", " ")

    data_limit_reached = False
    if "Max data limit reached" in message:
        data_limit_reached = True
        log.info("Max data limit reached, making new request for older data\n")

    if not message:
        log.info("No data found\n")
        message = []
    else:
        message = [
            msg[1].strip()
            for msg in re.findall("(//START)(.*?)(//END)", message, re.M | re.S)
        ]

    return message, data_limit_reached


def _format_period(period):
    days, hours, minutes = (
        period.days,
        period.seconds // 3600,
        (period.seconds // 60) % 60,
    )

    if minutes:
        return "now -%s minutes" % period.seconds / 60

    if hours:
        return "now -%s hours" % period.seconds / 3600

    if days:
        return "now -%s days" % days


def _format_time(timestamp):

    if isinstance(timestamp, basestring):
        if timestamp.startswith("P"):
            timestamp = isodate.parse_duration(timestamp)
        else:
            timestamp = isodate.parse_datetime(timestamp)

    if isinstance(timestamp, datetime):
        return timestamp.strftime("%Y/%j %H:%M:%S")
    elif isinstance(timestamp, timedelta):
        return _format_period(timestamp)


def _get_store_path(path, default_file_name):
    if path is None:
        path = os.path.join(util.get_ulmo_dir(), DEFAULT_FILE_PATH)

    if not os.path.exists(path):
        os.makedirs(path)

    return os.path.join(path, default_file_name)


def _parse(line):
    return {
        "dcp_address": line[:8],
        "message_timestamp_utc": datetime.strptime(line[8:19], "%y%j%H%M%S"),
        "failure_code": line[19:20],
        "signal_strength": line[20:22],
        "frequency_offset": line[22:24],
        "modulation_index": line[24:25],
        "data_quality_indicator": line[25:26],
        "goes_receive_channel": line[26:29],
        "goes_spacecraft_indicator": line[29:30],
        "uplink_carrier_status": line[30:32],
        "message_data_length": line[32:37],
        "dcp_message": line[37:],
    }



"""
    `NASA EARTHDATA ORNL DAAC Daymet`_ web services


    .. _NASA EARTHDATA ORNL DAAC Daymet: https://daymet.ornl.gov/dataaccess.html
"""
from __future__ import absolute_import

from . import core

from .core import get_daymet_singlepixel
from .core import get_variables

from tsgettoolbox.ulmo import util
